@@ -1,4 +1,4 @@
-# Copyright (C) 2006--2019  Kipp Cannon
+# Copyright (C) 2006--2020  Kipp Cannon
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -66,13 +66,9 @@ from . import types as ligolwtypes
 
 def get_array(xmldoc, name):
 	"""
-	Scan xmldoc for an array named name.  Raises ValueError if not
-	exactly 1 such array is found.
+	Deprecated.  use Array.get_array(xmldoc, name).
 	"""
-	arrays = Array.getArraysByName(xmldoc, name)
-	if len(arrays) != 1:
-		raise ValueError("document must contain exactly one %s array" % Array.ArrayName(name))
-	return arrays[0]
+	return Array.get_array(xmldoc, name)
 
 
 #
@@ -136,26 +132,25 @@ class ArrayStream(ligolw.Stream):
 		w(self.start_tag(indent))
 
 		array = self.parentNode.array
-		if array is not None:
+		if array is not None and array.size:
 			# avoid symbol and attribute look-ups in inner
 			# loop.  we use self.parentNode.shape to retrieve
 			# the array's shape, rather than just asking the
 			# array, to induce a sanity check that the Dim
 			# elements are correct for the array
 			linelen = self.parentNode.shape[0]
-			lines = array.size // linelen if array.size else 0
+			lines = array.size // linelen
 			tokens = iter(map(ligolwtypes.FormatFunc[self.parentNode.Type], array.T.flat))
 			islice = itertools.islice
 			join = self.Delimiter.join
 
-			if lines:
-				newline = u"\n" + indent + ligolw.Indent
+			newline = u"\n" + indent + ligolw.Indent
+			w(newline)
+			w(xmlescape(join(islice(tokens, linelen))))
+			newline = self.Delimiter + newline
+			for i in range(lines - 1):
 				w(newline)
 				w(xmlescape(join(islice(tokens, linelen))))
-				newline = self.Delimiter + newline
-				for i in range(lines - 1):
-					w(newline)
-					w(xmlescape(join(islice(tokens, linelen))))
 		w(u"\n" + self.end_tag(indent) + u"\n")
 
 
@@ -343,10 +338,30 @@ class Array(ligolw.Array):
 	def getArraysByName(cls, elem, name):
 		"""
 		Return a list of arrays with name name under elem.
+
+		See also .get_array().
 		"""
 		name = cls.ArrayName(name)
 		return elem.getElements(lambda e: (e.tagName == cls.tagName) and (e.Name == name))
 
+	@classmethod
+	def get_array(cls, xmldoc, name = None):
+		"""
+		Scan xmldoc for an array named name.  Raises ValueError if
+		not exactly 1 such array is found.  If name is None
+		(default), then the .arrayName attribute of this class is
+		used.  The Array class does not provide a .arrayName
+		attribute, but sub-classes could choose to do so.
+
+		See also .getArraysByName().
+		"""
+		if name is None:
+			name = cls.arrayName
+		elems = cls.getArraysByName(xmldoc, name)
+		if len(elems) != 1:
+			raise ValueError("document must contain exactly one %s array" % cls.ArrayName(name))
+		return elems[0]
+
 	#
 	# Element methods
 	#
